import logging

from collections import namedtuple
from typing import List, Union, Dict, Optional

from .api import event, task_event
from .appliance import min_ppa_version

logger = logging.getLogger(__name__)


class FileUpload(namedtuple("FileUpload", ("data", "filename", "size", "mtime", "type"))):
    """ Data returned by the input/file event.

    :param data: The file data.
    :param filename: The original filename.
    :param size: The size of the file in bytes.
    :param mtime: The last modified timestamp.
    :param type: A content type, if the browser was able to detect it.
    """

    def __repr__(self):
        return f"filename '{self.filename}', size '{self.size}', last_modified '{self.mtime}', type '{self.type}'"


@task_event(return_tuple=FileUpload)
def input_file(text: str, accept: Union[str, List[str]] = None, element_id: Optional[str] = None) -> FileUpload:
    """ Request the users uploads a file. The file uploaded will be made available to be downloaded by the user.

    :param text: The message to display when prompt for the file.
    :param accept: A mime type or extension to accept.
    :param element_id: The ID of the UI element. See :ref:`getting_started:Input & Output IDs`.

    :returns: A :class:`FileUpload` containing the file and its metadata.

    Example:

        .. code-block:: python

            # As the user to upload a configuration file
            config = ui.input_file(
                "Upload configuration",
                accept=".cfg",
            )
            print(config['data'])
    """
    return event("input/file", text, accept=accept, id=element_id)


@min_ppa_version("2.6.3")
@task_event(return_tuple=FileUpload)
def input_sensitive_file(
    text: str, accept: Union[str, List[str]] = None, element_id: Optional[str] = None
) -> FileUpload:
    """ Request the users uploads a sensitive file. This should be used when uploading a file that is sensitive
        such as: SSH Keys or Certificate Private Keys. The file uploaded will not be available to download to the user.

    :param text: The message to display when prompt for the file.
    :param accept: A mime type or extension to accept.
    :param element_id: The ID of the UI element. See :ref:`getting_started:Input & Output IDs`.

    :returns: A :class:`FileUpload` containing the file and its metadata.

    Example:

        .. code-block:: python

            # As the user to upload a configuration file
            config = ui.input_sensitive_file(
                "Upload configuration",
                accept=".cfg",
            )
            print(config['data'])
    """
    return event("input/sensitive_file", text, accept=accept, id=element_id)


@task_event()
def input_choice(text: str, *, options: List[Union[str, int]], element_id: Optional[str] = None) -> str:
    """ Display a dropdown box and wait until one value is submitted.

    :param text: The title of the dropdown.
    :param choices: A list of choices. Choices can be strings or integers.
    :param element_id: The ID of the UI element. See :ref:`getting_started:Input & Output IDs`.

    :returns: A list of choices. All choices are returned as strings, regardless
        of their type when supplied to this function.

    Example:

        .. code-block:: python

            choice = ui.input_choice("Please choose one value", options=["Option 1", "Option 2"])
    """
    return event("input/choice", text, options=list(options), id=element_id)


@task_event()
def input_table(text: str, *, header: List[str], data: List[list], element_id: Optional[str] = None) -> List[int]:
    """ Display a table and allow the user to select rows. Any number of rows can be
    selected.

    :param text: The title of the table.
    :param data: The table contents. This should be a list of lists, where each
        inner list is a table row containing cell value strings.
    :param header: The table column header. This should be a list of strings.
    :param element_id: The ID of the UI element. See :ref:`getting_started:Input & Output IDs`.

    :returns: A list of integers representing the table row indices selected by the user.

    Example:

        .. code-block:: python

            header = ["City", "Country"]
            data = [
                ["London", "England"],
                ["Paris", "France"],
                ["Rome", "Italy"],
                ["Cardiff", "Wales"],
            ]
            indices = ui.input_table(
                "Please choose some records",
                header=header,
                data=data,
            )
            selected_records = [data[i] for i in indices]
            ui.output_info(f"You selected {selected_records}")
    """
    return event("input/table", text, data=data, type="application/json", header=header, id=element_id)


@task_event()
def input_accept(text: str, element_id: Optional[str] = None) -> None:
    """ Display a button and block the task until it is clicked.

    :param text: The text to display inside the button.
    :param element_id: The ID of the UI element. See :ref:`getting_started:Input & Output IDs`.

    Example with supplied button text:

        .. code-block:: python

            ui.input_accept("Begin task")

    """
    return event("input/accept", text, id=element_id)


@task_event()
def input_confirm(
    text: str, confirm: Optional[str] = None, cancel: Optional[str] = None, element_id: Optional[str] = None
) -> bool:
    """ Display some text with configurable `confirm` and `cancel`
    buttons underneath.

    :param text: The text to display.
    :param confirm: The `confirm` button text. Defaults to `Confirm`.
    :param cancel: The `cancel` button text. Defaults to `Cancel`.
    :param element_id: The ID of the UI element. See :ref:`getting_started:Input & Output IDs`.

    :returns: True if confirm is pressed, False if cancel is pressed.

    Example with supplied button text:

        .. code-block:: python

            if ui.input_confirm(
                "Would you like to continue?",
                confirm="Yes",
                cancel="No"
            ):
                pass
            else:
                ui.output_error("Task will not continue")

    """
    return event("input/confirm", text, id=element_id, confirm=confirm, cancel=cancel)


@task_event()
def input_password(text: str, meter: bool = False, element_id: Optional[str] = None) -> str:
    """ Display a password input element. A password strength-meter
    will be displayed underneath the input field. The supplied password
    is delivered to the task but never stored in the database. The
    strength index is stored in the database for auditing purposes.

    :param text: The text to display with the input field.
    :param element_id: The ID of the UI element. See :ref:`getting_started:Input & Output IDs`.
    :param meter: Denotes whether a password meter should be shown in the PPA UI. Defaults to False.

    :returns: The supplied password as a string.

    Example:

        .. code-block:: python

            password = ui.input_password("New Password", meter=True)
            confirm_password = ui.input_password("Confirm Password")
            if password != confirm_password:
                ui.output_error("Passwords do not match.")

    """
    return event("input/password", text, id=element_id, meter=meter)


@task_event()
def input_text(
    text: str,
    default: Optional[str] = None,
    pattern: str = ".*",
    element_id: Optional[str] = None,
    required: bool = False,
) -> str:
    """ Display a free text input element.

    :param text: The text to display with the input field.
    :param default: The default value. Defaults to None.
    :param pattern: A regular expression used to validate the input.
    :param element_id: The ID of the UI element. See :ref:`getting_started:Input & Output IDs`.
    :param required: Denotes whether a value must be supplied. Defaults to False.

    :returns: The supplied text as a string.

    Example of a required field that must only contain one or more word characters:

        .. code-block:: python

            ui.input_text("Enter a word", pattern="\\w+", required=True)

    """
    return event("input/text", text, default=default, pattern=pattern, required=required, id=element_id)


@min_ppa_version("2.6.3")
@task_event()
def input_text_area(
    text: str,
    default: Optional[str] = None,
    pattern: str = ".*",
    element_id: Optional[str] = None,
    required: bool = False,
    rows: int = 5,
    max_rows: int = 50,
) -> str:
    """ Display a free multiline text input element.

    :param text: The text to display with the input field.
    :param default: The default value. Defaults to None.
    :param pattern: A regular expression used to validate the input.
    :param element_id: The ID of the UI element. See :ref:`getting_started:Input & Output IDs`.
    :param required: Denotes whether a value must be supplied. Defaults to False.
    :param rows: Denotes how many rows the input element should have.
    :param max_rows: Denotes the maximum number of rows the input element should expand to before
        forcing a scrollbar to appear.

    :returns: The supplied text as a string.

    Example of a required field that must only contain one or more word characters:

        .. code-block:: python

            ui.input_text_area("Enter a multiline comment", pattern="\\w+", required=True)

    """
    return event(
        "input/text",
        text,
        default=default,
        pattern=pattern,
        required=required,
        id=element_id,
        multiline=True,
        rows=rows,
        maxRows=max_rows,
    )


@task_event()
def output_markdown(doc: str, element_id: Optional[str] = None) -> Union[Dict, List]:
    """ Display a markdown document in the PPA task UI.

    :param doc: The markdown to display.
    :param element_id: The ID of the UI element. See :ref:`getting_started:Input & Output IDs`.

    Example:

        .. code-block:: python

            ui.output_markdown('''
            #### A Markdown Document

            Some example markdown text followed by a list.

            - Item One
            - Item Two
            - Item Three

            __Note: Some bold emphasis.__
            ''')

    .. seealso:: `Markdown CheatSheet <https://github.com/adam-p/markdown-here/wiki/Markdown-Cheatsheet#code>`_

    """
    return event("output/markdown", doc, id=element_id)


@task_event()
def output_error(text: str, element_id: Optional[str] = None) -> None:
    """ Display an error message in the PPA task UI.

    :param text: The error to display.
    :param element_id: The ID of the UI element. See :ref:`getting_started:Input & Output IDs`.

    Example:

        .. code-block:: python

            ui.output_error(
                "An operation failed. Please view the logs for more information"
            )
    """
    return event("output/error", text, id=element_id)


@task_event()
def output_info(text: str, element_id: Optional[str] = None) -> None:
    """ Display an info message in the PPA task UI.

    :param text: The info message to display.
    :param element_id: The ID of the UI element. See :ref:`getting_started:Input & Output IDs`.

    Example:

        .. code-block:: python

            ui.output_info("Operation complete")
    """
    return event("output/info", text, id=element_id)


@task_event()
def output_file(
    text: str, *, data: str, filename: str, type: str = "application/octet-stream", element_id: Optional[str] = None
) -> None:
    """ Offer a file download to the user.

    :param text: The message to display when offering the file.
    :param data: The file contents.
    :param filename: The name of the file.
    :param type: The content type. Defaults to `application/octet-stream`.
    :param element_id: The ID of the UI element. See :ref:`getting_started:Input & Output IDs`.

    Example:

        .. code-block:: python

            # download a hypothetical file 'report.csv'
            ui.output_file(
                "Download Report",
                data=report,
                filename="report.csv"
            )

    """
    return event("output/file", text, filename=filename, data=data, type=type, id=element_id)


@task_event()
def output_progress(text: str, *, percent: Union[int, None], element_id: Optional[str] = None) -> None:
    """ Display a progress bar.

    :param text: The title of the progress bar.
    :param percent: The current progress value. Defaults to None. If the supplied
        value is None, the progress bar will display as indeterminate.
    :param element_id: The ID of the UI element. See :ref:`getting_started:Input & Output IDs`.

    Example where an indeterminate progress bar is displayed, then updated
    after a time delay:

        .. code-block:: python

            ui.output_progress("Performing operation...", None, element_id="operation")
            time.sleep(10)
            ui.output_progress("Operation complete.", 100, element_id="operation")

    """
    return event("output/progress", text, percent=percent, id=element_id)


@task_event()
def output_table(text: str, *, header: List[str], data: List[list], element_id: Optional[str] = None) -> None:
    """ Display a table.

    :param text: The title of the table.
    :param data: The table contents. This should be a list of lists, where each
        inner list is a table row containing cell value strings.
    :param header: The table column header. This should be a list of strings.
    :param element_id: The ID of the UI element. See :ref:`getting_started:Input & Output IDs`.

    Example:

        .. code-block:: python

            header = ["City", "Country"]
            data = [
                ["London", "England"],
                ["Paris", "France"],
                ["Rome", "Italy"],
                ["Cardiff", "Wales"],
            ]
            ui.output_table(
                "Record Table",
                header=header,
                data=data,
            )
    """
    return event("output/table", text, data=data, type="application/json", header=header, id=element_id)
